/*********************************************************************************************************
* Software License Agreement (BSD License)                                                               *
* Author: Sebastien Decugis <sdecugis@nict.go.jp>							 *
*													 *
* Copyright (c) 2008, WIDE Project and NICT								 *
* All rights reserved.											 *
* 													 *
* Redistribution and use of this software in source and binary forms, with or without modification, are  *
* permitted provided that the following conditions are met:						 *
* 													 *
* * Redistributions of source code must retain the above 						 *
*   copyright notice, this list of conditions and the 							 *
*   following disclaimer.										 *
*    													 *
* * Redistributions in binary form must reproduce the above 						 *
*   copyright notice, this list of conditions and the 							 *
*   following disclaimer in the documentation and/or other						 *
*   materials provided with the distribution.								 *
* 													 *
* * Neither the name of the WIDE Project or NICT nor the 						 *
*   names of its contributors may be used to endorse or 						 *
*   promote products derived from this software without 						 *
*   specific prior written permission of WIDE Project and 						 *
*   NICT.												 *
* 													 *
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED *
* WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A *
* PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR *
* ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 	 *
* LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 	 *
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR *
* TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF   *
* ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.								 *
*********************************************************************************************************/

/* Logs and debug tracing facility.
 * 
 * This file contains the definitions for debug functions 
 *
 * The TRACE_DEBUG facility allow each file to report at separate levels:
 *  (default) INFO  -> to be able to follow execution path.
 *            FULL  -> all messages that are defined are printed.
 * 
 * Example: 
 * In the dictionary file, TRACE_LEVEL is defined to FULL before this file is included (or at compilation).
 * Then each message is reported through TRACE_DEBUG ( FULL, "message" ) or TRACE_DEBUG (INFO, "message").
 *
 */

#ifndef _LOG_H_
#define _LOG_H_

/* Logs */
#include <waaad/log-api.h> /* define the log_* functions */

/* The following additional functions are needed inside the daemon */

#define LOG_TO_SYSLOG 	0x1
#define LOG_TO_CONSOLE	0x2
#define LOG_TO_FILE	0x4

/*
 * FUNCTION:	log_init
 *
 * PARAMETERS:
 *  flags 	: Where the messages must be sent. combination of LOG_TO_* flags
 *
 * DESCRIPTION: 
 *  Initialize the log facility. 
 *
 * RETURN VALUE:
 *   0	: Application is now ready to log.
 *  !0  : An error occurred. 
 */
int log_init ( int flags );

/*
 * FUNCTION:	log_reinit
 *
 * PARAMETERS:
 *  flags 	: Where the messages must be sent. combination of LOG_TO_* flags
 *
 * DESCRIPTION: 
 *  Re-initialize the log facility. Basically, it gives the possibility to change the
 * destination of the messages. Will have no effect if the flags are not changed.
 *
 * RETURN VALUE:
 *   0	: Application is now ready to log to new set of facilities.
 *  !0  : An error occurred. 
 */
int log_reinit ( int flags );

/*
 * FUNCTION:	log_fini
 *
 * PARAMETERS:
 *  -
 *
 * DESCRIPTION: 
 *  Terminates the log facility. No log_* function must be called after this one.
 *
 * RETURN VALUE:
 *   0 : log facility closed properly.
 *  !0 : an error occurred (we may ignore it)
 */
int log_fini ( void );



/* The remaining of the file is dedicated to debug facility */

/* levels definitions */
#define NONE 0	/* Display no debug message */
#define INFO 1	/* Display errors only */
#define FULL 2  /* Display additional information to follow code execution */
#define FCTS 6  /* Display entry parameters of most functions */
#define CALL 9  /* Display calls to most functions (with CHECK macros) */

/* Default level is INFO */
#ifndef TRACE_LEVEL 
#define TRACE_LEVEL INFO
#endif /* TRACE_LEVEL */

/* The level of the module being compiled */
static int debug_level = TRACE_LEVEL;
/* We use an run-time variable to be able to change the level on-the-fly with a debugger */

/* The name of the current thread */
extern pthread_key_t thread_name;
void log_set_thread_name(char * name, char * cat);

/* Get a string describing current time */
char * log_time(char * buf, size_t len);

/* Macro to assign current thread name */
#define THREAD_NAME( name )		\
	log_set_thread_name( (name), NULL );

/* Macro to log debug */
#define TRACE_DEBUG(level,format,args... ) {										\
	if ( level <= debug_level + g_conf->log_verbose ) {								\
		char __buf[25];												\
		char * __thn = ((char *)pthread_getspecific(thread_name) ?: "unnamed");					\
		log_debug("\t |*%-30s\t%s\tin %s@%s:%d\n"								\
				"\t%s|%*s" format "\n",  								\
						__thn, log_time(__buf, sizeof(__buf)), __FUNCTION__, __FILE__, __LINE__,\
						(level < FULL)?"@":" ",level, "", ## args); 				\
	}														\
}
		
#define TRACE_ENTRY(_format,_args... ) \
	TRACE_DEBUG(FCTS, "->%s (" #_args ") = (" _format ") >", __FUNCTION__, ##_args );
	
#define TRACE_HERE()	\
	TRACE_DEBUG(INFO, " -- debug checkpoint -- ");
	
#define TRACE_DEBUG_ALL( str ) 	\
	TRACE_DEBUG(CALL, str );



#endif /* _LOG_H_ */
