/*********************************************************************************************************
* Software License Agreement (BSD License)                                                               *
* Author: Sebastien Decugis <sdecugis@nict.go.jp>							 *
*													 *
* Copyright (c) 2008, WIDE Project and NICT								 *
* All rights reserved.											 *
* 													 *
* Redistribution and use of this software in source and binary forms, with or without modification, are  *
* permitted provided that the following conditions are met:						 *
* 													 *
* * Redistributions of source code must retain the above 						 *
*   copyright notice, this list of conditions and the 							 *
*   following disclaimer.										 *
*    													 *
* * Redistributions in binary form must reproduce the above 						 *
*   copyright notice, this list of conditions and the 							 *
*   following disclaimer in the documentation and/or other						 *
*   materials provided with the distribution.								 *
* 													 *
* * Neither the name of the WIDE Project or NICT nor the 						 *
*   names of its contributors may be used to endorse or 						 *
*   promote products derived from this software without 						 *
*   specific prior written permission of WIDE Project and 						 *
*   NICT.												 *
* 													 *
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED *
* WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A *
* PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR *
* ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 	 *
* LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 	 *
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR *
* TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF   *
* ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.								 *
*********************************************************************************************************/

/* Peers facility.
 *
 * This file contains the functions to manage the _peer_t structures and the peer hash table.
 *
 */
#include "waaad-internal.h"
#include "peer-internal.h"


/* The hash table */
_peer_hash_t 	_peer_hash[1 << _PEER_HASH_SIZE];

/* Initialize the hash table */
int _peer_struct_init()
{
	int i;
	
	/* Initialize the hash table */
	memset(_peer_hash, 0, sizeof(_peer_hash));
	
	for (i = 0; i < sizeof(_peer_hash) / sizeof(_peer_hash[0]); i++) {
		uti_list_init(&_peer_hash[i].all, 	&_peer_hash[i]);
		uti_list_init(&_peer_hash[i].actives, 	&_peer_hash[i]);
		CHECK_POSIX(  pthread_mutex_init(&_peer_hash[i].lock, NULL)  );
	}
	
	return 0;
}

/* Destroy the hash table and all peers */
int _peer_struct_fini()
{
	int i;
	
	for (i = 0; i < sizeof(_peer_hash) / sizeof(_peer_hash[0]); i++) {
		while ( ! IS_LIST_EMPTY( &_peer_hash[i].all )  ) {
			_peer_t * peer = _P(_peer_hash[i].all.next->o);
			CHECK_FCT_DO(_peer_psm_stop(peer, -1), /* continue */ );
			CHECK_FCT_DO(_peer_struct_unlink(peer), /* continue */ );
			CHECK_FCT_DO(_peer_struct_destroy(&peer), /* continue */ );
		}
		CHECK_POSIX_DO(  pthread_mutex_destroy(&_peer_hash[i].lock), /* continue */  );
	}
	
	/* done */
	return 0;
}

/* Add an element in a peer_appl_t list of a peer */
int _peer_struct_appl_add( _peer_t * peer, application_id_t a, vendor_id_t v, peer_app_type_t t)
{
	int i;
	
	if (a == 0xffffffff) {
		peer->p_app_relay = 1;
		return 0;
	}
	
	CHECK_MALLOC( peer->p_app_list = (peer_appl_t *)realloc(peer->p_app_list, (peer->p_app_size + 2) * sizeof(peer_appl_t)) );
	peer->p_app_size += 1;

	/* We must move the elements after the one we are inserting */
	/* ( a b c d e 0 ) --> ( a b N c d e 0 ) */

	memset( &peer->p_app_list[peer->p_app_size], 0, sizeof(peer_appl_t));

	i = peer->p_app_size - 2;

	while ((i >= 0) && (peer->p_app_list[i].a > a)) {
		memcpy( &peer->p_app_list[i + 1], &peer->p_app_list[i], sizeof(peer_appl_t));
		i--;
	}
	peer->p_app_list[i+1].a = a;
	peer->p_app_list[i+1].v = v;
	peer->p_app_list[i+1].t = t;
	
	return 0;
}
void _peer_struct_appl_reset( _peer_t * peer )
{
	free(peer->p_app_list);
	peer->p_app_list = NULL;
	peer->p_app_size = 0;
	peer->p_app_relay = 0;
	return;
}

/* Add an element in the p_remotesec list of a peer */
int _peer_struct_secid_add( _peer_t * peer, uint32_t secid)
{
	CHECK_MALLOC( peer->p_remotesec = (uint32_t *)realloc(peer->p_remotesec, (peer->p_remotesec_sz + 1) * sizeof(uint32_t)) );
	peer->p_remotesec[peer->p_remotesec_sz] = secid;
	peer->p_remotesec_sz += 1;
	return 0;
}
void _peer_struct_secid_reset( _peer_t * peer )
{
	free(peer->p_remotesec);
	peer->p_remotesec = NULL;
	peer->p_remotesec_sz = 0;
	return;
}
int _peer_struct_secid_check( _peer_t * peer, uint32_t secid)
{
	int i;
	
	for (i=0; i<peer->p_remotesec_sz; i++)
		if (peer->p_remotesec[i] == secid)
			return 1;
	
	return 0;
}

/* reset the remote addresses published in CER/CEA */
void _peer_struct_addrs_reset( _peer_t * peer )
{
	peer->p_peeraddr_sz = 0;
	free(peer->p_peeraddr);
	peer->p_peeraddr = NULL;
}

/* reset security module to default nosec */
void _peer_struct_secmod_reset(_peer_t * peer)
{
	if (peer->p_sec_hdl) {
		CHECK_FCT_DO(  sec_modunlink(peer->p_sec_hdl),  /* nothing */  );
		peer->p_sec_hdl = NULL;
	}
	
	peer->p_secmod = &_peer_secini_mod;
	
	while (!IS_LIST_EMPTY( &peer->p_sec_list ) )
	{
		uti_list_t * li = peer->p_sec_list.next;
		uti_list_unlink(li);
		CHECK_FCT_DO( sec_modunlink(((_sec_item_t *)(li))->hdl), /* continue */ );
		free(li);
	}
	
	return;
}

/* Create a new peer structure. It is not saved in the hash table */
/* id can be NULL or lower-case string */
int _peer_struct_malloc(char * id, _peer_t **new) 
{
	int ret = 0;
	char * diamid;
	_peer_t * peer;
	
	/* Ensure the location for the new peer is valid */
	CHECK_PARAMS( new );
	*new = NULL;
	
	/* The diameter id of the new peer */
	if (id) {
		diamid = id;
	} else {
		CHECK_MALLOC( diamid = strdup("(unknown peer)")  );
	}
	
	/* Now create the structure and initialize it */
	CHECK_MALLOC_DO( peer = _P(malloc(sizeof(_peer_t))), { ret = ENOMEM; goto rb1; } );
	
	/* Set everything to 0 */
	memset(peer, 0, sizeof(_peer_t));
	
	/* Initialize the fields */
	peer->p_eyec = PEER_EYEC;
	
	/* The hash will be computed later, only if the peer has to be inserted into the hash list */
	uti_list_init(&peer->p_global, peer);
	
	peer->p_diamid = diamid;
	/* The realm is NULL yet, it will be set later */
	peer->p_hbh = lrand48();
	
	/* p_flags are set to 0 */
	peer->p_state = STATE_DISABLED;
	uti_list_init(&peer->p_active, peer);
	/* p_psm will be created when the PSM is started */
	CHECK_POSIX_DO(  ret = pthread_mutex_init(&peer->p_lock, NULL),  goto rb2  );
	CHECK_POSIX_DO(  ret = pthread_cond_init(&peer->p_condvar, NULL),  goto rb3  );
	/* p_ts will be set later */
	uti_list_init(&peer->p_events, NULL);
	
	CHECK_FCT_DO(  meq_new(&peer->p_in_q),  goto rb4  );
	/* p_in_th will be created by the psm */
	CHECK_FCT_DO(  meq_new(&peer->p_out_q),  goto rb5  );
	/* p_out_th will be created by the psm */
	uti_list_init(&peer->p_sent, NULL);
	
	/* p_addinfo is set later */
	uti_list_init(&peer->p_exp_list, peer);
	
	/* All connection information will be set later */
	
	uti_list_init(&peer->p_sec_list, NULL);
	/* p_sec_hdl: no security handler initially */
	peer->p_secmod = &_peer_secini_mod;
	/* p_sec_session is initialized later when sending/receiving starts */
	
	/* The remaining of the structure is not set yet */
	
	/* Done: return 0 */
	*new = peer;
	return 0;
	
rb5:
	CHECK_FCT_DO(  meq_del(&peer->p_in_q), /* nothing */  );
rb4:
	CHECK_POSIX_DO(  pthread_cond_destroy(&peer->p_condvar), /* continue */  );
rb3:
	CHECK_POSIX_DO(  pthread_mutex_destroy(&peer->p_lock), /* continue */  );
rb2:
	free(peer);
rb1:
	free(diamid);
	return ret;
}

/* Remove a peer structure from all global lists. The peer must be in DISABLED state. */
int _peer_struct_unlink(_peer_t * peer)
{
	/* First, remove from expiry or deleted list if needed */
	if ( ! IS_LIST_EMPTY(&peer->p_exp_list) ) {
		CHECK_POSIX( pthread_mutex_lock(&_peers_lists_mtx) );
		uti_list_unlink(&peer->p_exp_list);
		CHECK_POSIX( pthread_mutex_unlock(&_peers_lists_mtx) );
	}
	
	/* Then remove from the hash table */
	if ( ! IS_LIST_EMPTY(&peer->p_global) ) {
		ASSERT( peer->p_hash ); /* Theoritically, the hash value may really be 0... Well, this is for debug, the ASSERT will not be compiled in final release. */
		CHECK_POSIX( pthread_mutex_lock(H_LOCK(peer->p_hash)) );
		uti_list_unlink(&peer->p_global);
		H_COUNT(peer->p_hash) -= 1;
		CHECK_POSIX( pthread_mutex_unlock(H_LOCK(peer->p_hash)) );
	}
	
	/* Done */
	return 0;
}

/* Destroy a peer structure. It must be already unlinked from any global list and in DISABLED state. Its lock must not be held. */
int _peer_struct_destroy(_peer_t ** peer)
{
	_peer_t * p;
	
	/* Check the parameter is acceptable */
	CHECK_PARAMS( peer && VALIDATE_PEER(*peer) );
	p=*peer;
	CHECK_PARAMS( (p->p_state == STATE_DISABLED) && IS_LIST_EMPTY(&p->p_global)  );
	
	/* Erase the location of the peer in the calling space */
	*peer = NULL;
	
	/* Some sanity checks */
	ASSERT( IS_LIST_EMPTY(&p->p_active)  );
	ASSERT( IS_LIST_EMPTY(&p->p_exp_list)  );
	ASSERT( p->p_psm == (pthread_t)NULL );
	ASSERT( p->p_ext_session == NULL );
	
	/* Now destroy the peer data */
	
	/* Free the strings */
	free(p->p_diamid);
	free(p->p_realm);
	
	/* Destroy the mutex and condvar */
	CHECK_POSIX_DO( pthread_mutex_destroy(&p->p_lock), /* continue */ );
	CHECK_POSIX_DO( pthread_cond_destroy(&p->p_condvar), /* continue */ );
	
	/* Destroy any pending event */
	_peer_events_reset(p);
	
	/* Handle any message that would remain in this peer */
	CHECK_FCT_DO( _peer_failover(p), /* continue */  );
	
	/* Destroy the message queues */
	CHECK_FCT_DO(  meq_del(&p->p_in_q), /* continue */  );
	CHECK_FCT_DO(  meq_del(&p->p_out_q), /* continue */  );
	
	/* Destroy the initial information of the peer */
	if (p->p_addinfo.pa_flags & PA_FREE)
		free(p->p_addinfo.pa_cbparm);
	
	/* Destroy the socket if not done yet */
	if (p->p_sock) {
		CHECK_SYS_DO( shutdown(p->p_sock, SHUT_RDWR), /* continue */ );
	}
	
	/* Destroy the array of endpoints from CER/CEA */
	free(p->p_peeraddr);
	
	/* Destroy security information */
	while ( ! IS_LIST_EMPTY( &p->p_sec_list ) ) {
		uti_list_t * next = p->p_sec_list.next;
		uti_list_unlink(next);
		free(next);
	}
	
	if (p->p_sec_hdl) {
		CHECK_FCT_DO( sec_modunlink(p->p_sec_hdl), /* continue */ );
	}
	
	free( p->p_remotesec );
	
	/* List of applciations */
	free( p->p_app_list );
	
	/* Product name */
	free( p->p_prodname );
	
	/* Election data if any */
	if (p->p_electdata) {
		(void)_peer_struct_destroy(&p->p_electdata->peer);
		free(p->p_electdata);
	}
	
	/* Finally, free the peer itself */
	free(p);
	
	return 0;
}

/* Finding a peer from a diameter id. diamid must be lowercase! 
 * Upon return, the hash line is locked and must be unlocked.
 * If item is found, it is returned in peer. otherwise, the next element is returned.
 */
int _peer_struct_search(char * diamid, size_t diamidlen, uint32_t * hash, uti_list_t ** peerglob)
{
	uint32_t h;
	size_t   l;
	uti_list_t *li;
	int found = 0;
	
	TRACE_ENTRY("%p %d %p %p", diamid, diamidlen, hash, peerglob);
	
	if (diamidlen == 0)
		l = strlen(diamid);
	else
		l = diamidlen;
	
	h = uti_hash(diamid, l);
	if (hash)
		*hash = h;
	
	/* Now search the peer in the appropriate hash line */
	CHECK_POSIX(  pthread_mutex_lock( H_LOCK(h) )  );
	
	/* Search the line for this peer. A line is ordered by hash and diamid */
	for (li = H_ALL(h)->next; li != H_ALL(h); li = li->next) {
		int cmp = 0;
		_peer_t * cur = _P(li->o);
		
		if (cur->p_hash < h)
			continue;
		
		if (cur->p_hash > h)
			break;
		
		cmp = strncmp(cur->p_diamid, diamid, l);
		
		if (cmp < 0)
			continue;
		
		if (cmp == 0)
			found = 1;
		
		break;
	}
	
	/* return the results */
	if (peerglob)
		*peerglob = li;
	
	return found ? 0 : ENOENT;
}


